{-# OPTIONS --no-coverage-check #-} 

module Renaming where
open import Utils
open import Term
open import Value
open import Normal
open import BigStep
open import Recursive

infixl 50 _•_ 

data Ren (Γ : Cn) : Cn -> Set where
  r : Ren Γ ε
  _<<_ : forall {Δ}{σ} -> Ren Γ Δ -> Var Γ σ -> Ren Γ (Δ , σ)

mapr : forall {Γ}{Γ'}{Δ} -> (forall {σ} -> Var Γ σ -> Var Γ' σ) -> Ren Γ Δ -> 
       Ren Γ' Δ
mapr f r = r
mapr f (vs << v) = mapr f vs << f v  

idr : forall {Γ} -> Ren Γ Γ
idr {ε} = r 
idr {Γ , σ} = mapr (\v -> vS v) (idr {Γ})  << vZ 

renv : forall {Γ}{Δ}{σ} -> Var Δ σ -> Ren  Γ Δ  -> Var Γ σ
renv vZ     (_ << v)   = v 
renv (vS v) (vs << _)  = renv v vs 

_•_ : forall {B}{Γ}{Δ} -> Ren Γ Δ -> Ren B Γ -> Ren B Δ
(vs << v) • vs' = vs • vs' << renv v vs'
r • vs = r

mutual
  rennf : forall {Γ}{Δ}{σ} -> Nf Δ σ -> Ren  Γ Δ  -> Nf Γ σ
  rennf (λn n)  rs = λn (rennf n (mapr (\v -> vS v) rs << vZ))
  rennf (nen n) rs = nen (rennef n rs)
  rennf Zn      _  = Zn
  rennf (Sn n)  rs = Sn (rennf n rs)

  rennef : forall {Γ}{Δ}{σ} -> Ne Nf Δ σ -> Ren  Γ Δ  -> Ne Nf Γ σ
  rennef (var v)    rs = var (renv v rs)
  rennef (app n n') rs = app (rennef n rs) (rennf n' rs)
  rennef (npr n f z) rs = npr (rennef n rs) (rennf f rs) (rennf z rs)

mutual
  renval : forall {Γ}{Δ}{σ} -> Val Δ σ -> Ren  Γ Δ  -> Val Γ σ
  renval (λv t vs) rs = λv t (renenv vs rs)
  renval (nev n)   rs = nev (rennev n rs)
  renval  Zv       _  = Zv
  renval (Sv v)    rs = Sv (renval v rs)
 
  renenv : forall {B}{Γ}{Δ} -> Env Γ Δ -> Ren B Γ -> Env B Δ
  renenv e        rs = e
  renenv (vs ◂ v) rs = renenv vs rs ◂ renval v rs 

  rennev : forall {Γ}{Δ}{σ} -> Ne Val Δ σ -> Ren  Γ Δ  -> Ne Val Γ σ
  rennev (var v)   rs = var (renv v rs) 
  rennev (app n v) rs = app (rennev n rs) (renval v rs)
  rennev (npr n f z) rs = npr (rennev n rs) (renval f rs) (renval z rs) 

rnwk : forall Γ Δ -> Ren (Γ ++ Δ) Γ
rnwk Γ ε = idr {Γ}
rnwk Γ (Δ , σ) = mapr (\v -> vS v)  (rnwk Γ Δ)

wknf++ : forall {Γ}{σ} -> Nf Γ σ -> (Δ : Cn) -> Nf (Γ ++ Δ) σ
wknf++ {Γ} n Δ = rennf n (rnwk Γ Δ) 

wknef++ : forall {Γ}{σ} -> Ne Nf Γ σ -> (Δ : Cn) -> Ne Nf (Γ ++ Δ) σ
wknef++ {Γ} n Δ = rennef n (rnwk Γ Δ) 

mutual
  ren[] : forall {B}{Γ}{Δ}{σ}{t : Tm Δ σ}{vs : Env Γ Δ}{v : Val Γ σ} ->
          t at vs ⇓ v -> (rs : Ren B Γ) -> t at renenv vs rs ⇓ renval v rs
  ren[] rø             rs = rø 
  ren[] rλ             rs = rλ 
  ren[] (r$ p p' p'')  rs = r$ (ren[] p rs) (ren[] p' rs) (ren$ p'' rs ) 
  ren[] (r[] p p')     rs = r[] (ren[]ˢ p rs) (ren[] p' rs) 
  ren[] rZ             _  = rZ
  ren[] (rS p)         rs = rS (ren[] p rs)
  ren[] (rp p q p' q') rs = 
    rp (ren[] p rs) (ren[] q rs) (ren[] p' rs) (renpr⇓ q' rs)

  renpr⇓ : forall {Γ}{Δ}{σ}{n : Val Δ N}{f : Val Δ (N → σ → σ)}{z : Val Δ σ} ->
           {v : Val Δ σ} -> pr⇓ n f z v -> (rs : Ren Γ Δ) ->
           pr⇓ (renval n rs) (renval f rs) (renval z rs) (renval v rs)
  renpr⇓ rpZ            _  = rpZ
  renpr⇓ (rpS p p' p'') rs = rpS (ren$ p rs) (renpr⇓ p' rs) (ren$ p'' rs)  

  ren$ : forall {Γ}{Δ}{σ}{τ}{f : Val Δ (σ → τ)}{a : Val Δ σ}{v : Val Δ τ} ->
         f $ a ⇓ v -> (rs : Ren Γ Δ) -> renval f rs $ renval a rs ⇓ renval v rs
  ren$ (r$λ p) rs = r$λ (ren[] p rs) 
  ren$ r$n     rs = r$n 

  ren[]ˢ : forall {A}{B}{Γ}{Δ}{ts : Tmˢ Γ Δ}{vs : Env B Γ} -> 
           {vs' : Env B Δ} -> ts atˢ vs ⇓ vs' -> (rs : Ren A B) ->
           ts atˢ renenv vs rs ⇓ renenv vs' rs
  ren[]ˢ rid       rs = rid 
  ren[]ˢ (r◃ p p') rs = r◃ (ren[]ˢ p rs)(ren[] p' rs)  
  ren[]ˢ r↑        _  = r↑
  ren[]ˢ (r○ p p') rs = r○ (ren[]ˢ p rs) (ren[]ˢ p' rs) 
 
-- to define renaming for quote derivations we need to move between weakenings
-- and renamings for the higher order case.

lemVar : forall {Γ}{Δ}{σ}(v : Var Δ σ)(τ : Ty)(rs : Ren Γ Δ) ->
  renv (vS v) (mapr (\v -> vS v) rs << vZ) == vS (renv v rs)
lemVar vZ     τ (_ << v)  = refl⁼ 
lemVar (vS v) τ (vs << _) = lemVar v τ vs 

mutual
  lemVal : forall {Γ}{Δ}{σ}(v : Val Δ σ)(τ : Ty)(rs : Ren Γ Δ) ->
    renval (wkv τ v) (mapr (\v -> vS v) rs << vZ) == wkv τ (renval v rs)
  lemVal (nev n)   τ rs = resp (\n -> nev n) (lemNeVal n τ rs) 
  lemVal (λv t vs) τ rs = resp (\vs -> λv t vs) (lemEnv vs τ rs) 
  lemVal Zv        _ _  = refl⁼
  lemVal (Sv v)    τ rs = resp (\v -> Sv v) (lemVal v τ rs)  

  lemNeVal : forall {Γ}{Δ}{σ}(n : Ne Val Δ σ)(τ : Ty)(rs : Ren Γ Δ) ->
    rennev (wkn τ n) (mapr (\v -> vS v) rs << vZ) == wkn τ (rennev n rs)
  lemNeVal (var v) τ rs = resp (\v -> var {Val} v) (lemVar v τ rs) 
  lemNeVal (app n v) τ rs = resp2 (\n v -> app n v)
                                  (lemNeVal n τ rs)
                                  (lemVal v τ rs)
  lemNeVal (npr n f z) τ rs = resp3 (\n f z -> npr n f z)
                                    (lemNeVal n τ rs)
                                    (lemVal f τ rs)
                                    (lemVal z τ rs) 

  lemEnv : forall {B}{Γ}{Δ}(vs : Env Γ Δ)(τ : Ty)(rs : Ren B Γ) ->
    renenv (wke τ vs) (mapr (\v -> vS v) rs << vZ) == wke τ (renenv vs rs)
  lemEnv e        τ rs = refl⁼ 
  lemEnv (vs ◂ v) τ rs = resp2 (\vs v -> vs ◂ v)
                               (lemEnv vs τ rs)
                               (lemVal v τ rs)

mutual
  renq : forall {Γ}{Δ}{σ}{v : Val Δ σ}{n : Nf Δ σ} -> q v ⇓ n -> 
    (rs : Ren Γ Δ) -> q (renval v rs) ⇓ (rennf n rs)
  renq (q→ {._} {σ} {τ} {f} {v} p q) rs = 
    q→ y (renq q (mapr (\v -> vS v) rs << vZ)) where
      x = ren$ p (mapr (\v -> vS v) rs << vZ) 
      y = help$ (lemVal f σ rs) x
  renq (qn p)   rs = qn (renqn p rs) 
  renq qZ       _  = qZ
  renq (qS p)   rs = qS (renq p rs)

  renqn : forall {Γ}{Δ}{σ}{v : Ne Val Δ σ}{n : Ne Nf Δ σ} -> qⁿ v ⇓ n ->
    (rs : Ren Γ Δ)  -> qⁿ (rennev v rs)  ⇓ (rennef n rs)
  renqn qv       rs = qv 
  renqn (qa p q) rs = qa (renqn p rs ) (renq q rs) 

-- now I need to prove that wknf++ is the same as weakening by a renaming

renwkVar : forall {Γ}{Δ}{σ}(v : Var Δ σ)(τ : Ty)(rs : Ren Γ Δ) ->
  vS (renv v rs) == renv v (mapr (\v -> vS v) rs)
renwkVar vZ     τ (_ << v)  = refl⁼ 
renwkVar (vS v) τ (vs << _) = renwkVar v τ vs 

mutual
  renwkVal : forall {Γ}{Δ}{σ}(v : Val Δ σ)(τ : Ty)(rs : Ren Γ Δ) ->
    wkv τ (renval v rs) == renval v (mapr (\v -> vS v) rs)
  renwkVal (nev n)   τ vs  = resp (\n -> nev n) (renwkNeVal n τ vs) 
  renwkVal (λv t vs) τ vs' = resp (\vs -> λv t vs) (renwkEnv vs τ vs') 

  renwkEnv : forall {B}{Γ}{Δ}(vs : Env Γ Δ)(τ : Ty)(rs : Ren B Γ) ->
    wke τ (renenv vs rs) == renenv vs (mapr (\v -> vS v) rs)
  renwkEnv e τ vs = refl⁼ 
  renwkEnv (vs ◂ v) τ vs' = resp2 (\vs v -> vs ◂ v)
                                  (renwkEnv vs τ vs')
                                  (renwkVal v  τ vs')

  renwkNeVal : forall {Γ}{Δ}{σ}(n : Ne Val Δ σ)(τ : Ty) ->
    (rs : Ren Γ Δ) ->
    wkn τ (rennev n rs) == rennev n (mapr (\v -> vS v) rs)
  renwkNeVal (var v)   τ vs = resp (\v -> var {Val} v) (renwkVar v τ vs) 
  renwkNeVal (app n v) τ vs = resp2 (\n v -> app n v) 
                                     (renwkNeVal n τ vs)
                                     (renwkVal   v τ vs)
  renwkNeVal (npr n f z) τ vs = resp3 (\n f z -> npr n f z) 
                                      (renwkNeVal n τ vs) 
                                      (renwkVal f τ vs)
                                      (renwkVal z τ vs) 

renidVar : forall {Γ}{σ}(v : Var Γ σ) ->  renv v idr == v
renidVar {ε} ()
renidVar {Γ , τ} vZ     = refl⁼
renidVar {Γ , τ} (vS v) = trans⁼ (sym⁼ (renwkVar v  τ  idr)) 
                                 (resp (\v -> vS v {τ}) (renidVar v))  

mutual 
  renidVal : forall {Γ}{σ}(v : Val Γ σ) ->  renval v idr == v
  renidVal (nev n) = resp (\n -> nev n) (renidNeVal n)
  renidVal (λv t vs) = resp (\vs -> λv t vs) (renidEnv vs)

  renidNeVal : forall {Γ}{σ}(n : Ne Val Γ σ) ->  rennev n idr == n
  renidNeVal (var v) = resp (\v -> var {Val} v) (renidVar v)
  renidNeVal (app n v) = resp2 (\n v -> app n v)
                               (renidNeVal n)
                               (renidVal v)
  renidNeVal (npr n f z) = resp3 (\n f z -> npr n f z) 
                                 (renidNeVal n)
                                 (renidVal f)
                                 (renidVal z) 

  renidEnv : forall {Γ}{Δ}(vs : Env Γ Δ) ->  renenv vs idr == vs
  renidEnv e = refl⁼
  renidEnv (vs ◂ v) = resp2 (\vs v -> vs ◂ v) 
                            (renidEnv vs)
                            (renidVal v)

lemwk++renˣ : forall {Γ}{σ}(v : Var Γ σ) Δ ->
              renv v (rnwk Γ Δ) == wkvar++ Δ v
lemwk++renˣ v ε = renidVar v 
lemwk++renˣ {Γ} v (Δ , σ) with wkvar++ Δ v | lemwk++renˣ v Δ
... | ._ | refl⁼ = sym⁼ (renwkVar v σ (rnwk Γ Δ)) 

lemwk++ren : forall {Γ}{σ}(v : Val Γ σ)(Δ : Cn) ->
             renval v (rnwk Γ Δ) == wkv++ Δ v
lemwk++ren v ε        = renidVal v 

lemwk++ren {Γ} v (Δ , τ) = trans⁼ (sym⁼ (renwkVal v  τ (rnwk Γ Δ ))) 
                             (resp (\v -> wkv τ v) (lemwk++ren v Δ)) 

lemwk++renⁿ : forall {Γ}{σ}(n : Ne Val Γ σ)(Δ : Cn) ->
              rennev n (rnwk Γ Δ) == wkn++ Δ n
lemwk++renⁿ n ε        = renidNeVal n 
lemwk++renⁿ {Γ} n (Δ , τ) = trans⁼ (sym⁼ (renwkNeVal n  τ (rnwk Γ Δ ))) 
                             (resp (\n -> wkn τ n) (lemwk++renⁿ n Δ)) 

wkqn : forall {Γ}{σ}{n : Ne Val Γ σ}{n' : Ne Nf Γ σ} -> qⁿ n ⇓ n' ->
  (Δ : Cn) -> qⁿ (wkn++ Δ n) ⇓ (wknef++ n' Δ)
wkqn {Γ} {σ} {n} p Δ = helpqn (lemwk++renⁿ n Δ) (renqn p (rnwk Γ Δ))  

wkq : forall {Γ}{σ}{v : Val Γ σ}{n : Nf Γ σ} -> q v ⇓ n -> (Δ : Cn) ->
  q (wkv++ Δ v) ⇓ (wknf++ n Δ)
wkq {Γ} {σ} {v} p Δ = helpq (lemwk++ren v Δ) (renq p (rnwk Γ Δ)) 

r2s : forall {Γ}{Δ} -> Ren Γ Δ -> Tmˢ Γ Δ
r2s (vs << v) = r2s vs ◃ ⌜ v ⌝ˣ
r2s {Γ} r = mkid Γ

mutual
  rent : forall {Γ}{Δ}{σ} -> Tm Δ σ -> Ren Γ Δ -> Tm Γ σ
  rent t rs = t [ r2s rs ]

  rents : forall {B}{Γ}{Δ} -> Tmˢ Γ Δ -> Ren B Γ -> Tmˢ B Δ
  rents ts rs = ts ○ r2s rs  

lem⌜⌝Var : forall {Γ}{Δ}{σ}(v : Var Δ σ)(rs : Ren Γ Δ) ->
           rent ⌜ v ⌝ˣ rs ≡ ⌜ renv v rs ⌝ˣ
lem⌜⌝Var vZ     (rs << v) = ø≡
lem⌜⌝Var (vS v) (rs << v') = 
  trans (trans [][]≡ ([]≡ refl ↑◃≡ˢ)) (lem⌜⌝Var v rs) 

mutual 
  lem⌜⌝Val : forall {Γ}{Δ}{σ}(v : Val Δ σ)(rs : Ren Γ Δ) ->
             rent ⌜ v ⌝ʷ rs ≡ ⌜ renval v rs ⌝ʷ
  lem⌜⌝Val (nev n)   rs = lem⌜⌝NeVal n rs
  lem⌜⌝Val (λv t vs) rs = trans [][]≡ ([]≡ refl (lem⌜⌝Env vs rs)) 

  lem⌜⌝NeVal : forall {Γ}{Δ}{σ}(n : Ne Val Δ σ)(rs : Ren Γ Δ) ->
               rent ⌜ n ⌝ʷⁿ rs ≡ ⌜ rennev n rs ⌝ʷⁿ
  lem⌜⌝NeVal (var v)   rs = lem⌜⌝Var v rs 
  lem⌜⌝NeVal (app n v) rs = trans $[]≡ ($≡ (lem⌜⌝NeVal n rs) (lem⌜⌝Val v rs )) 
  lem⌜⌝NeVal (npr n f z) rs = 
    trans pr[]≡ (pr≡ (lem⌜⌝NeVal n rs) (lem⌜⌝Val f rs) (lem⌜⌝Val z rs)) 


  lem⌜⌝Env : forall {B}{Γ}{Δ}(vs : Env Γ Δ)(rs : Ren B Γ) ->
             rents ⌜ vs ⌝ˢ  rs ≡ˢ ⌜ renenv vs rs ⌝ˢ
  lem⌜⌝Env e r = id≡ˢ
  lem⌜⌝Env e (rs << v) = transˢ (transˢ ○○≡ˢ (○≡ reflˢ ↑◃≡ˢ)) (lem⌜⌝Env e rs)
  lem⌜⌝Env (vs ◂ v) rs = transˢ ○◃≡ˢ (◃≡ (lem⌜⌝Env vs rs) (lem⌜⌝Val v rs)) 

-- I need to prove that a weakening subsitution is that same as 
-- weakening a renaming

ren↑ : forall {Γ}{Δ}(rs : Ren Γ Δ)(σ : Ty) -> 
  r2s rs ○ ↑ σ ≡ˢ r2s (mapr (\v -> vS v) rs)
ren↑ r         σ = reflˢ 
ren↑ (rs << v) σ = transˢ ○◃≡ˢ (◃≡ (ren↑ rs σ) refl)

mutual
  lem⌜⌝Nf : forall {Γ}{Δ}{σ}(n : Nf Δ σ)(rs : Ren Γ Δ) ->
     rent ⌜ n ⌝ rs ≡ ⌜ rennf n rs ⌝
  lem⌜⌝Nf (λn n)  rs =  
    trans λ[]≡ (λ≡ (trans ([]≡ refl (◃≡ (ren↑ rs _) refl))
                          (lem⌜⌝Nf n (mapr (\v -> vS v) rs << vZ)))) 
  lem⌜⌝Nf (nen n) rs = lem⌜⌝NeNf n rs 

  lem⌜⌝NeNf : forall {Γ}{Δ}{σ}(n : Ne Nf Δ σ)(rs : Ren Γ Δ) ->
     rent ⌜ n ⌝ⁿ rs ≡ ⌜ rennef n rs ⌝ⁿ
  lem⌜⌝NeNf (var v)    rs = lem⌜⌝Var v rs 
  lem⌜⌝NeNf (app n n') rs = trans $[]≡ ($≡ (lem⌜⌝NeNf n rs) (lem⌜⌝Nf n' rs)) 
  lem⌜⌝NeNf (npr n f z) rs = 
    trans pr[]≡ (pr≡ (lem⌜⌝NeNf n rs) (lem⌜⌝Nf f rs) (lem⌜⌝Nf z rs))  

ren⌜⌝ : forall {Γ}{Δ}{σ}{t t' : Tm Δ σ} -> t ≡ t' -> (rs : Ren Γ Δ) -> 
        rent t rs ≡ rent t' rs 
ren⌜⌝ p rs = []≡ p  reflˢ  

help⌜⌝ : forall {Γ}{σ}{v v' : Val Γ σ} -> v == v' -> {n : Nf Γ σ} ->
  ⌜ v ⌝ʷ ≡ ⌜ n ⌝ ->   ⌜ v' ⌝ʷ ≡ ⌜ n ⌝
help⌜⌝ refl⁼ p = p

wk⌜⌝ : forall {Γ}{σ}{v : Val Γ σ}{n : Nf Γ σ} -> ⌜ v ⌝ʷ ≡ ⌜ n ⌝ -> (Δ : Cn) ->
        ⌜ wkv++ Δ v ⌝ʷ ≡ ⌜ wknf++ n Δ ⌝
wk⌜⌝ {Γ} {σ} {v} {n} p Δ = help⌜⌝ (lemwk++ren v Δ) {wknf++ n Δ} 
   (trans (sym (lem⌜⌝Val v (rnwk Γ Δ))) 
          (trans (ren⌜⌝ p (rnwk Γ Δ) ) (lem⌜⌝Nf n (rnwk Γ Δ))))

help⌜⌝ⁿ : forall {Γ}{σ}{n n' : Ne Val Γ σ} -> n == n' -> {n'' : Ne Nf Γ σ} ->
          ⌜ n ⌝ʷⁿ  ≡ ⌜ n'' ⌝ⁿ -> ⌜ n' ⌝ʷⁿ ≡ ⌜ n'' ⌝ⁿ
help⌜⌝ⁿ refl⁼ p = p

wk⌜⌝ⁿ : forall {Γ}{σ}{n : Ne Val Γ σ}{n' : Ne Nf Γ σ} -> ⌜ n ⌝ʷⁿ ≡ ⌜ n' ⌝ⁿ ->
        (Δ : Cn) -> ⌜ wkn++ Δ n ⌝ʷⁿ ≡ ⌜ wknef++ n' Δ ⌝ⁿ
wk⌜⌝ⁿ {Γ} {σ} {n} {n'} p Δ = help⌜⌝ⁿ (lemwk++renⁿ n Δ) {wknef++ n' Δ}
   (trans (sym (lem⌜⌝NeVal n (rnwk Γ Δ))) 
          (trans (ren⌜⌝ p (rnwk Γ Δ)) (lem⌜⌝NeNf n' (rnwk Γ Δ)))) 

-- I think this probably has almost the same structure as the recursive
-- functions so it's no surprise it doesn't termination check...

mutual 
  lem$$++ : forall {Γ}{σ τ}(f : Val Γ (σ → τ))(a : Val Γ σ) Δ ->
            wkv++ Δ f $$ wkv++ Δ a == wkv++ Δ (f $$ a)
  lem$$++ (λv t vs) a Δ with wkv++ Δ (ev t (vs ◂ a)) | sym⁼ (lemev++ t (vs ◂ a) Δ)
  ... | ._ | refl⁼ with wke++ Δ (vs ◂ a) | lem◂ Δ vs a 
  ... | ._ | refl⁼ with wkv++ Δ (λv t vs) | lemλwk t vs Δ
  ... | ._ | refl⁼ = refl⁼  
  lem$$++ (nev n)   a Δ with wkv++ Δ (nev n) | lemNwk n Δ
  ... | ._ | refl⁼ with wkv++ Δ (nev (app n a)) | lemNwk (app n a) Δ
  ... | ._ | refl⁼ with wkn++ Δ (app n a) | lemappwk n a Δ 
  ... | ._ | refl⁼ = refl⁼ 

  lemev++ : forall {B Δ}{σ}(t : Tm Δ σ)(vs : Env B Δ) Γ ->
            ev t (wke++ Γ vs) == wkv++ Γ (ev t vs)
  lemev++ ø (vs ◂ v) Δ with wke++ Δ (vs ◂ v) | lem◂ Δ vs v
  ... | ._  | refl⁼ = refl⁼ 
  lemev++ (λ t) vs Δ with wkv++ Δ (λv t vs) | lemλwk t vs Δ 
  ... | ._ | refl⁼ = refl⁼
  lemev++ (t $ u) vs Δ with wkv++ Δ (ev t vs $$ ev u vs) | lem$$++ (ev t vs) (ev u vs) Δ 
  ... | ._ | refl⁼ with wkv++ Δ (ev t vs) | lemev++ t vs Δ 
  ... | ._ | refl⁼ with wkv++ Δ (ev u vs) | lemev++ u vs Δ 
  ... | ._ | refl⁼ = refl⁼
  lemev++ (t [ ts ]) vs Δ with wkv++ Δ (ev t (evˢ ts vs)) | lemev++ t (evˢ ts vs) Δ 
  ... | ._ | refl⁼ with wke++ Δ (evˢ ts vs) | lemevˢ++ ts vs Δ
  ... | ._ | refl⁼  = refl⁼
  lemev++ {Γ} Z vs Δ with wkv++ Δ (Zv {Γ}) | lemZwk {Γ} Δ
  ... | ._ | refl⁼ = refl⁼
  lemev++ (S n) vs Δ with wkv++ Δ (Sv (ev n vs)) | lemSwk (ev n vs) Δ
  ... | ._ | refl⁼ with wkv++ Δ (ev n vs) | lemev++ n vs Δ 
  ... | ._ | refl⁼ = refl⁼
  lemev++ (pr n f z) vs Δ with wkv++ Δ (ev (pr n f z) vs) | lemev++ (pr n f z) vs Δ 
  ... | ._ | refl⁼ = refl⁼

  lemevˢ++ : forall {A} {Γ Δ}(ts : Tmˢ Γ Δ)(vs : Env A Γ) B ->
           evˢ ts (wke++ B vs) == wke++ B (evˢ ts vs) 
  lemevˢ++ id vs Δ = refl⁼
  lemevˢ++ (ts ◃ t) vs Δ with wke++ Δ (evˢ ts vs ◂ ev t vs) | lem◂ Δ (evˢ ts vs) (ev t vs)
  ... | ._ | refl⁼  with wke++ Δ (evˢ ts vs) | lemevˢ++ ts vs Δ 
  ... | ._ | refl⁼ with wkv++ Δ (ev t vs) | lemev++ t vs Δ
  ... | ._ | refl⁼ = refl⁼
  lemevˢ++ (↑ σ) (vs ◂ v) Δ with wke++ Δ (vs ◂ v) | lem◂ Δ vs v
  ... | ._ | refl⁼ = refl⁼
  lemevˢ++ (ts ○ us) vs Δ with wke++ Δ (evˢ ts (evˢ us vs)) | lemevˢ++ ts (evˢ us vs) Δ
  ... | ._ | refl⁼ with wke++ Δ (evˢ us vs) | lemevˢ++ us vs Δ 
  ... | ._ | refl⁼ = refl⁼  

-- This is sufficient for ev but if we want to push weakenings past quote we 
-- need renaming for the whole lot (this could really be subsumed by the next
-- thing).

mutual
  lemevren : forall {B Γ Δ}{σ}(t : Tm Δ σ)(vs : Env Γ Δ)(rs : Ren B Γ) ->
             ev t (renenv vs rs) == renval (ev t vs) rs
  lemevren ø (vs ◂ v) rs = refl⁼ 
  lemevren (λ t) vs rs = refl⁼ 
  lemevren (t $ u) vs rs with ev t (renenv vs rs) | sym⁼ (lemevren t vs rs)
  ... | ._ | refl⁼ with ev u (renenv vs rs) | sym⁼ (lemevren u vs rs)
  ... | ._ | refl⁼ = lem$$ren (ev t vs) (ev u vs) rs 
  lemevren (t [ ts ]) vs rs with evˢ ts (renenv vs rs) | sym⁼ (lemevˢren ts vs rs)
  ... | ._ | refl⁼ = lemevren t (evˢ ts vs) rs 
  lemevren Z vs rs = refl⁼ 
  lemevren (S n) vs rs = resp (\v -> Sv v) (lemevren n vs rs) 
  lemevren (pr n f z) vs rs with ev n (renenv vs rs) | sym⁼ (lemevren n vs rs)
  ... | ._ | refl⁼  with ev f (renenv vs rs) | sym⁼ (lemevren f vs rs)
  ... | ._ | refl⁼  with ev z (renenv vs rs) | sym⁼ (lemevren z vs rs)
  ... | ._ | refl⁼ = lemprimvren (ev n vs) (ev f vs) (ev z vs) rs 

  lemevˢren : forall {A B Γ Δ}(ts : Tmˢ Γ Δ)(vs : Env B Γ)(rs : Ren A B) ->
             evˢ ts (renenv vs rs) == renenv (evˢ ts vs) rs
  lemevˢren id vs rs = refl⁼ 
  lemevˢren (ts ◃ t) vs rs = resp2 (\vs v -> vs ◂ v) (lemevˢren ts vs rs) (lemevren t vs rs) 
  lemevˢren (↑ σ) (vs ◂ v) rs = refl⁼ 
  lemevˢren (ts ○ us) vs rs with evˢ us (renenv vs rs) | sym⁼ (lemevˢren us vs rs)
  ... | ._ | refl⁼ = lemevˢren ts (evˢ us vs) rs 
 
  lemprimvren : forall {Γ Δ}{σ}(n : Val Δ N)(f : Val Δ (N → σ → σ))(z : Val Δ σ)(rs : Ren Γ Δ) ->
             primv (renval n rs) (renval f rs) (renval z rs) == renval (primv n f z) rs
  lemprimvren Zv f z rs = refl⁼ 
  lemprimvren (Sv n) f z rs with renval (f $$ n $$ primv n f z) rs | lem$$ren (f $$ n) (primv n f z) rs
  ... | ._ | refl⁼ with renval (f $$ n) rs | lem$$ren f n rs 
  ... | ._ | refl⁼ with renval (primv n f z) rs | lemprimvren n f z rs
  ... | ._ | refl⁼ = refl⁼ 
  lemprimvren (nev n) f z rs = refl⁼ 

  lem$$ren : forall {Γ Δ}{σ τ}(f : Val Δ (σ → τ))(a : Val Δ σ)(rs : Ren Γ Δ) ->
             renval f rs $$ renval a rs == renval (f $$ a) rs
  lem$$ren (λv t vs) a rs = lemevren t (vs ◂ a) rs 
  lem$$ren (nev n) a rs = refl⁼

mutual 
  lemquoteren : forall {Γ Δ}{σ}(v : Val Δ σ)(rs : Ren Γ Δ) ->
    rennf (quote v) rs == quote (renval v rs)
  lemquoteren {σ = σ → τ} f rs with wkv σ (renval f rs) | lemVal f σ rs
  ... | ._ | refl⁼ with rennf (quote (wkv σ f $$ nev (var vZ))) (mapr (\v -> vS v) rs << vZ) | lemquoteren {σ = τ} (wkv σ f $$ nev (var vZ)) (mapr (\v -> vS v) rs << vZ) 
  ... | ._ | refl⁼ with renval (wkv σ f $$ nev (var vZ)) (mapr (\v -> vS v) rs << vZ) | lem$$ren (wkv σ f) (nev (var vZ)) (mapr (\v -> vS v) rs << vZ)
  ... | ._ | refl⁼ = refl⁼
  lemquoteren {σ = N} Zv rs = refl⁼ 
  lemquoteren {σ = N} (Sv n) rs = resp (\n -> Sn n) (lemquoteren n rs) 
  lemquoteren {σ = N} (nev n) rs = resp (\n -> nen n) (lemquoteⁿren n rs)  

  lemquoteⁿren : forall {Γ Δ}{σ}(n : Ne Val Δ σ)(rs : Ren Γ Δ) ->
    rennef (quoteⁿ n) rs == quoteⁿ (rennev n rs)
  lemquoteⁿren (var v) rs = refl⁼ 
  lemquoteⁿren (app n v) rs = resp2 (\n v -> app n v)
                                    (lemquoteⁿren n rs)
                                    (lemquoteren v rs) 
  lemquoteⁿren (npr n f z) rs = resp3 (\n f z -> npr n f z)
                                       (lemquoteⁿren n rs)
                                       (lemquoteren f rs)
                                       (lemquoteren z rs) 

lemeqⁿ++ : forall {Γ}{σ}(n n' : Ne Val Γ σ) Δ -> 
          quoteⁿ n == quoteⁿ n' -> 
          quoteⁿ (wkn++ Δ n) == quoteⁿ (wkn++ Δ n')
lemeqⁿ++ {Γ} n n' Δ p with wkn++ Δ n | lemwk++renⁿ n Δ 
... | ._ | refl⁼ with wkn++ Δ n' | lemwk++renⁿ n' Δ
... | ._ | refl⁼ with quoteⁿ (rennev n (rnwk Γ Δ)) | lemquoteⁿren n (rnwk Γ Δ) 
... | ._ | refl⁼ with quoteⁿ (rennev n' (rnwk Γ Δ)) | lemquoteⁿren n' (rnwk Γ Δ)
... | ._ | refl⁼ with quoteⁿ n | p
... | ._ | refl⁼ = refl⁼ 
